from .validator import Validator  # noqa: F401 # imported but unused
from .required_validator import RequiredValidator  # noqa: F401 # imported but unused
from .value_type_validator import ValueTypeValidator  # noqa: F401 # imported but unused



from concrete_settings.exceptions import ValidationError
from concrete_settings.types import Undefined
from .validator import Validator


class RequiredValidator(Validator):
    def __init__(self, message: str = None):
        if message is None:
            message = (
                'Setting `{name}` is required to have a value. '
                'Current value is `Undefined`'
            )
        self.message = message

    def __call__(self, value, *, name, owner, **ignore):
        if value == Undefined:
            msg = self.message.format(name=name)
            raise ValidationError(msg)



import os
import sys
from sphinx.pycode.parser import Parser


def extract_doc_comments_from_class_or_module(module_name, cls_name=None):
    # read the contents of the module which contains the settings
    # and parse it via Sphinx parser
    module_path = getattr(sys.modules[module_name], '__file__', None)
    if not module_path or not os.path.exists(module_path):
        return {}

    with open(module_path, 'r') as f:
        module_code = f.read()
    parser = Parser(module_code)
    parser.parse()
    return parser.comments


def extract_docstrings_from_code(code):
    parser = Parser(code)
    parser.parse()



import sys
from typing import Any


class UndefinedMeta(type):
    def __bool__(self):
        return False

    def __str__(self):
        return 'Undefined value'


class Undefined(metaclass=UndefinedMeta):
    def __new__(cls, *args, **kwargs):
        raise RuntimeError(f'{cls} should not be instantiated')


class GuessSettingType:
    #: Recognized Setting value types
    KNOWN_TYPES = [
        bool,  # bool MUST come before int, as e.g. isinstance(True, int) == True
        int,
        float,
        complex,
        list,
        tuple,
        range,
        bytes,
        str,
        frozenset,
        set,
        dict,
    ]

    @classmethod
    def guess_type_hint(cls, val):
        if val is Undefined:
            return Any

        for t in cls.KNOWN_TYPES:
            if isinstance(val, t):
                return t
        return Any


def type_hints_equal(hint1, hint2):
    """Returns true if type hints equal.

    For example:

        type_hints_equal(typing.List, list) => True
    """
    assert hint1 is not None
    assert hint2 is not None

    if hint1 is hint2:
        return True

    h1 = getattr(hint1, '__origin__', hint1) or hint1  # `... or hint 1`
    h2 = getattr(hint2, '__origin__', hint2) or hint2  # - protect from `None` value

    if h1 is h2:
        return True

    PYTHON_36 = (3, 6)
    if (sys.version_info.major, sys.version_info.minor) == PYTHON_36:
        h1 = getattr(hint1, '__extra__', hint1)
        h2 = getattr(hint2, '__extra__', hint2)

    return h1 is h2



from .yaml_source import YamlSource  # noqa: F401 # imported but unused
from .json_source import JsonSource  # noqa: F401 # imported but unused
from .envvar_source import EnvVarSource  # noqa: F401 # imported but unused
from .python_source import PythonSource  # noqa: F401 # imported but unused



from typing import TYPE_CHECKING

from typing_extensions import Protocol


if TYPE_CHECKING:
    from ..settings import Setting, Settings


class Validator(Protocol):
    """A validator is a callable that raises an exception if a value is wrong.

    A validator accepts a value as a mandatory argument, and keyword-only arguments
    referring to settings, setting and setting's name."""

    def __call__(
        self,
        value,
        *,
        name: str = None,
        owner: 'Settings' = None,
        setting: 'Setting' = None,
    ):
        """Validate a value. Raise `ValidationError` if value is wrong."""
        ...



"""Built-in Settings update strategies.

This module contains the basic and default update strategies used
when calling `concrete_settings.Settings.update`
"""
import abc

from typing_extensions import Protocol


class Strategy(Protocol):
    @abc.abstractmethod
    def __call__(self, current_value, new_value):
        ...


def overwrite(current_val, new_value):
    return new_value


def append(current_val, new_value):
    return current_val + new_value


def prepend(current_value, new_value):
    return new_value + current_value


default: Strategy = overwrite



import importlib.util
from pathlib import Path
from typing import Any, Tuple, Union, Type

from concrete_settings.sources import FileSource, register_source, NotFound


@register_source
class PythonSource(FileSource):
    extensions = ['.py']

    def __init__(self, path):
        super().__init__(path)
        self._data = None

    def read(self, setting, parents: Tuple[str, ...] = ()) -> Union[Type[NotFound], Any]:
        if self._data is None:
            self._data = self._read_file(self.path)

        d = self._data
        for key in parents:
            d = d[key]

        val = d.get(setting.name, NotFound)
        return val

    @staticmethod
    def _read_file(path: str):
        module_name = Path(path).stem
        spec = importlib.util.spec_from_file_location(module_name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore
        return vars(module)



import types
from typing import Any, Union

from ..settings import Setting, PropertySetting, Settings


class GenericBehaviorMeta(type):
    def __call__(cls, *args, init_with_arguments=False, **kwargs):
        # Explicitly tell the metaclass that
        # Behavior.__init__(arg1, arg2, ...)
        # is expected
        # Case: @validate(validator1, ...)

        if init_with_arguments:
            return super().__call__(*args, **kwargs)

        # Act as a decorator
        _decorating_setting_or_method = (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], (Setting, types.FunctionType))
        )

        if _decorating_setting_or_method:
            if isinstance(args[0], types.FunctionType):
                setting = PropertySetting(args[0])
            else:
                setting = args[0]

            bhv = super().__call__()
            return bhv(setting)
        else:
            bhv = super().__call__(*args, **kwargs)
            return bhv

    def __rmatmul__(cls, setting: Any):
        if not isinstance(setting, Setting):
            setting = Setting(setting)

        bhv = cls()
        return bhv(setting)


class BehaviorWithArgumentsMeta(GenericBehaviorMeta):
    def __call__(cls, *args, **kwargs):
        return super().__call__(*args, **kwargs, init_with_arguments=True)


class Behavior(metaclass=GenericBehaviorMeta):
    def __call__(self, setting_or_method: Union[Setting, types.FunctionType]):
        setting: Setting

        if isinstance(setting_or_method, types.FunctionType):
            setting = PropertySetting(setting_or_method)
        else:
            setting = setting_or_method
        # Act as a decorator
        self.decorate(setting)
        return setting

    def __rmatmul__(self, setting: Any):
        if not isinstance(setting, Setting):
            setting = Setting(setting)

        self.decorate(setting)
        return setting

    def decorate(self, setting: Setting):
        pass


class GetterSetterBehavior(Behavior):
    def decorate(self, setting: Setting):
        # replace setting.get_value and .set_value by
        # corresponding behavior methods
        self._setting_get_value = setting.get_value
        self._setting_set_value = setting.set_value
        setting.get_value = types.MethodType(self.get_value, setting)  # type: ignore
        setting.set_value = types.MethodType(self.set_value, setting)  # type: ignore
        super().decorate(setting)

    def get_value(self, setting: Setting, owner: Settings) -> Any:
        return self._setting_get_value(owner)

    def set_value(self, setting: Setting, owner: Settings, value):
        self._setting_set_value(owner, value)



import functools
import logging
import types
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

from .docreader import extract_doc_comments_from_class_or_module
from .exceptions import StructureError, ValidationError, ValidationErrorDetails
from .sources import get_source, AnySource, Source, NotFound
from .sources.strategies import Strategy, default as default_update_strategy
from .types import Undefined, GuessSettingType, type_hints_equal
from .validators import Validator, ValueTypeValidator

logger = logging.getLogger(__name__)

INVALID_SETTINGS = '__invalid__settings__'


# ==== Settings classes ==== #
# ========================== #
class Setting:
    value: Any
    type_hint: Any
    validators: Tuple[Validator, ...]
    name: str
    override: bool

    def __init__(
        self,
        value: Any = Undefined,
        *,
        doc: str = '',
        validators: Tuple[Validator, ...] = (),
        type_hint: Any = GuessSettingType,
        override: bool = False,
    ):
        self.value = value
        self.type_hint = type_hint
        self.validators = tuple(validators)
        self.__doc__ = doc
        self.name = ""
        self.override = override

    def __set_name__(self, _, name):
        self.name = name

    def __get__(
        self, owner: Optional['Settings'], owner_type=None
    ) -> Union[Any, 'Setting']:
        assert isinstance(
            owner, (type(None), Settings)
        ), "owner should be None or an instance of Settings"

        # == class-level access ==
        if not owner:
            return self

        # == object-level access ==
        return self.get_value(owner)

    def get_value(self, owner):
        return getattr(owner, f"__setting_{self.name}_value", self.value)

    def __set__(self, owner: 'Settings', val):
        assert isinstance(owner, Settings), "owner should be an instance of Settings"
        self.set_value(owner, val)

    def set_value(self, owner: 'Settings', val):
        setattr(owner, f"__setting_{self.name}_value", val)


class PropertySetting(Setting):
    def __init__(self, *args, **kwargs):
        decorating_without_arguments = (
            len(args) == 1 and len(kwargs) == 0 and callable(args[0])
        )
        if decorating_without_arguments:
            self._init_decorator_without_arguments(args[0])
        else:
            if args != ():
                raise TypeError(
                    'No positional arguments should be passed to '
                    f'{self.__class__.__name__}.__init__()'
                )
            if 'value' in kwargs:
                raise TypeError(
                    '"value" keyword argument should not be passed to '
                    f'{self.__class__.__name__}.__init__()'
                )

            self._init_decorator_with_arguments(**kwargs)

    def _init_decorator_without_arguments(self, fget: Callable):
        super().__init__()
        self.__call__(fget)

    def _init_decorator_with_arguments(self, **kwargs):
        super().__init__(value=Undefined, **kwargs)
        self.fget = None

    def __call__(self, fget: Callable):
        # since functools.update_wrapper overwrite self.__doc__
        # we have to temporarily persist it
        doc_before_wrapping = self.__doc__

        functools.update_wrapper(self, fget)

        # Restore __doc__
        self.__doc__ = self.__doc__ or doc_before_wrapping

        # Extract type_hint from fget annotations if needed
        if self.type_hint is GuessSettingType:
            self.type_hint = fget.__annotations__.get('return', self.type_hint)

        self.fget = fget
        return self

    def __get__(
        self, owner: Optional['Settings'], owner_type=None
    ) -> Union[Any, 'Setting']:
        assert isinstance(
            owner, (type(None), Settings)
        ), "owner should be None or an instance of Settings"

        # == class-level access ==
        if not owner:
            return self

        if self.fget is None:
            raise AttributeError("Unreadable attribute")

        return self.get_value(owner)

    def get_value(self, owner: 'Settings'):
        return self.fget(owner)

    def __set__(self, owner: 'Settings', val):
        raise AttributeError("Can't set attribute: property setting cannot be set")


class SettingsMeta(type):
    def __new__(mcs, name, bases, class_dict):
        new_dict = mcs.class_dict_to_settings(class_dict, bases)
        mcs._add_settings_help(name, new_dict)
        return super().__new__(mcs, name, bases, new_dict)

    @classmethod
    def class_dict_to_settings(mcs, class_dict: dict, bases: List[type]):
        new_dict = {}
        annotations = class_dict.get("__annotations__", {})

        for name, attr in class_dict.items():
            new_attr = attr

            # Make a Setting out of ALL_UPPERCASE_ATTRIBUTE
            if (
                not isinstance(attr, Setting)
                and mcs._is_setting_name(name)
                and mcs._can_be_converted_to_setting_automatically(attr)
            ):
                new_attr = mcs._make_setting_from_attribute(name, attr, annotations)

            # Should we try to guess a type_hint for a Setting?
            if isinstance(new_attr, Setting) and new_attr.type_hint is GuessSettingType:
                new_attr.type_hint = mcs._guess_type_hint(
                    name, new_attr, annotations, bases
                )

            new_dict[name] = new_attr

        return new_dict

    @classmethod
    def _make_setting_from_attribute(
        mcs, name, attr, annotations
    ) -> Union[PropertySetting, Setting]:
        # is it a class method?
        if isinstance(attr, types.FunctionType):
            return PropertySetting(attr)

        type_hint = annotations.get(name, GuessSettingType)
        return Setting(attr, doc="", type_hint=type_hint)

    @classmethod
    def _guess_type_hint(mcs, name, setting: Setting, annotations, bases: List[type]):
        # we still have to check annotations,
        # e.g. if the setting was instantiated by behavior
        annotation_type_hint = annotations.get(name, GuessSettingType)
        if annotation_type_hint is not GuessSettingType:
            return annotation_type_hint

        # try to get the type hint from the base classes
        for base in bases:
            try:
                base_type_hint = getattr(base, name).type_hint
                return base_type_hint
            except AttributeError:
                pass

        guessed_setting_type = GuessSettingType.guess_type_hint(setting.value)
        return guessed_setting_type

    @classmethod
    def _is_setting_name(mcs, name: str) -> bool:
        """Return True if name is written in the upper case"""
        return not name.startswith('_') and name.upper() == name

    @classmethod
    def _can_be_converted_to_setting_automatically(mcs, attr: Any) -> bool:
        """Return False if attribute should not be converted
           to a Setting automatically"""
        callable_types = (property, classmethod, staticmethod)
        return not isinstance(attr, callable_types)

    @classmethod
    def _add_settings_help(mcs, cls_name: str, class_dict: dict):
        if '__module__' not in class_dict:
            # class is not coming from a module
            return

        settings = {
            name: attr for name, attr in class_dict.items() if isinstance(attr, Setting)
        }
        if not settings:
            # class seems to contain to settings
            return

        if all(setting.__doc__ for setting in settings.values()):
            # All settings of the class have been explicitly documented.
            # Since explicit documentation overrides comment-docs,
            # there is no need to proceed further
            return

        # read the contents of the module which contains the settings
        # and parse it via Sphinx parser
        cls_module_name = class_dict['__module__']

        comments = extract_doc_comments_from_class_or_module(cls_module_name, cls_name)

        for name, setting in settings.items():
            if setting.__doc__:
                # do not modify an explicitly-made setting documentation
                continue

            comment_key = (cls_name, name)
            try:
                setting.__doc__ = comments[comment_key]
            except KeyError:
                # no comment-style documentation exists
                pass


class Settings(Setting, metaclass=SettingsMeta):
    default_validators: Tuple[Validator, ...] = ()
    mandatory_validators: Tuple[Validator, ...] = (ValueTypeValidator(),)

    _is_being_validated: bool

    _errors: ValidationErrorDetails = {}

    def __init__(self, **kwargs):
        assert (
            'value' not in kwargs
        ), '"value" argument should not be passed to Settings.__init__()'
        assert (
            'type_hint' not in kwargs
        ), '"type_hint" argument should not be passed to Settings.__init__()'

        super().__init__(value=self, type_hint=self.__class__, **kwargs)

        self._is_being_validated = False
        self._verify_structure()

    def _verify_structure(self):
        # verify whether the setting on Nth level of the inheritance hierarchy
        # corresponds to the setting on N-1th level of the hierarchy.
        for name, classes in self._get_settings_classes().items():
            for c0, c1 in zip(classes, classes[1:]):
                # start with setting object of the first classes
                s0 = c0.__dict__[name]
                s1 = c1.__dict__[name]
                differences = self._settings_diff(s0, s1)
                if differences:
                    diff = '; '.join(differences)
                    raise StructureError(
                        f'in classes {c0} and {c1} setting {name} has'
                        f' the following difference(s): {diff}'
                    )

    def _get_settings_classes(self) -> Dict[str, List[Type['Settings']]]:
        # _settings_classes is helper list which can be used in
        # settings reading and validation routines.
        # 1. Iterate through __mro__ classes in reverse order - so that
        #    iteration happens from the most-base class to the current one.
        # 2. Store found settigns as {name: [cls, ...]} to settings_classes
        settings_classes: Dict[str, List[Type['Settings']]] = defaultdict(list)

        assert self.__class__.__mro__[-3] is Settings

        # __mro__[:-2] - skip Settings and object bases
        for cls in reversed(self.__class__.__mro__[:-3]):
            for attr, val in cls.__dict__.items():
                if isinstance(val, Setting):
                    settings_classes[attr].append(cls)
        return dict(settings_classes)

    def _settings_diff(self, s0: Setting, s1: Setting) -> List[str]:
        NO_DIFF = []  # type: ignore
        differences = []

        # No checks are performed if setting is overridden
        if s1.override:
            return NO_DIFF

        if not type_hints_equal(s0.type_hint, s1.type_hint):
            differences.append(f'types differ: {s0.type_hint} != {s1.type_hint}')

        return differences

    @classmethod
    def settings_attributes(cls) -> Iterator[Tuple[str, Setting]]:
        for name in dir(cls):
            attr = getattr(cls, name)
            if isinstance(attr, Setting):
                yield name, attr

    def is_valid(self, raise_exception=False) -> bool:
        self._errors = {}
        self._errors = self._run_validation(raise_exception)
        return self._errors == {}

    def _run_validation(self, raise_exception=False) -> ValidationErrorDetails:
        self._is_being_validated = True
        errors = {}

        # validate each setting individually
        for name, setting in self.settings_attributes():
            setting_errors = self._validate_setting(name, setting, raise_exception)
            if setting_errors:
                errors[name] = setting_errors

        if errors == {}:
            try:
                self.validate()
            except ValidationError as e:
                if raise_exception:
                    raise e
                else:
                    errors[INVALID_SETTINGS] = [str(e)]

        self._is_being_validated = False
        return errors

    def _validate_setting(
        self, name: str, setting: Setting, raise_exception=False
    ) -> ValidationErrorDetails:
        value: Setting = getattr(self, name)

        errors: List[ValidationErrorDetails] = []
        validators = setting.validators or self.default_validators
        validators += self.mandatory_validators

        for validator in validators:
            try:
                validator(value, name=name, owner=self, setting=setting)
            except ValidationError as e:
                if raise_exception:
                    raise ValidationError({name: e.details}) from e
                errors.append(str(e))
            except Exception as e:
                if raise_exception:
                    raise ValidationError({name: str(e)}) from e
                else:
                    errors.append(str(e))

        # nested Settings
        if isinstance(value, Settings):
            nested_settings = value
            try:
                nested_settings.is_valid(raise_exception=raise_exception)
            except ValidationError as e:
                assert raise_exception
                e.prepend_source(name)
                raise ValidationError({name: e.details}) from e

            if nested_settings.errors:
                errors.append(nested_settings.errors)

        return errors

    def validate(self):
        pass

    def update(self, source: AnySource, strategies: dict = None):
        strategies = strategies if strategies is not None else {}
        assert isinstance(strategies, Mapping), '`strategies` type should be `dict`'

        source_obj = get_source(source)
        self._update(self, source_obj, parents=(), strategies=strategies)

    def _update(
        self,
        settings: 'Settings',
        source: Source,
        parents: Tuple[str, ...] = (),
        strategies: Dict[str, Strategy] = None,
    ):
        """Recursively update settings object from dictionary"""
        strategies = strategies or {}

        for name, setting in settings.settings_attributes():
            if isinstance(setting, Settings):
                settings._update(setting, source, (*parents, name), strategies)
            else:
                full_setting_name = f'{".".join(parents) and "."}{name}'

                if full_setting_name in strategies:
                    update_strategy = strategies[full_setting_name]
                    logger.debug(
                        'Updating setting %s with strategy %s',
                        full_setting_name,
                        getattr(update_strategy, '__qualname__', 'unknown strategy'),
                    )
                else:
                    update_strategy = default_update_strategy

                update_to_val = source.read(setting, parents)
                if update_to_val is NotFound:
                    continue

                current_val = getattr(settings, name)
                new_val = update_strategy(current_val, update_to_val)
                setattr(settings, name, new_val)

    def extract_to(self, destination: Union[types.ModuleType, dict], prefix: str = ''):
        if prefix != '':
            prefix = prefix + '_'

        if isinstance(destination, types.ModuleType):
            destination = destination.__dict__

        for name, attr in self.settings_attributes():
            var_name = prefix + name
            if isinstance(attr, Settings):  # nested settings
                attr.extract_to(destination, var_name)
            else:
                destination[var_name] = getattr(self, name)

    @property
    def errors(self) -> ValidationErrorDetails:
        return self._errors

    @property
    def is_being_validated(self) -> bool:
        return self._is_being_validated



from typing import Dict, List, Tuple, Optional

from concrete_settings import Settings, Undefined

# This is defined here as a do-nothing function because we can't import
# django.utils.translation -- that module depends on the settings.
from concrete_settings.contrib.behaviors import required


# flake8: noqa


def gettext_noop(s):
    return s


class Django30Settings(Settings):
    DEBUG: bool = False

    #: Whether the framework should propagate raw exceptions rather than catching
    #: them. This is useful under some testing situations and should never be used
    #: on a live site.
    DEBUG_PROPAGATE_EXCEPTIONS: bool = False

    #: People who get code error notifications.
    #: In the format [
    #     ('Full Name', 'email@example.com'),
    #     ('Full Name', 'anotheremail@example.com')
    # ]
    ADMINS: List[Tuple[str, str]] = []

    #: List of IP addresses, as strings, that:
    #:   * See debug comments, when DEBUG is true
    #:   * Receive x-headers
    INTERNAL_IPS: List = []

    #: Hosts/domain names that are valid for this site.
    #: "*" matches anything, ".example.com" matches example.com and all subdomains
    ALLOWED_HOSTS: List = []

    #: Local time zone for this installation. All choices can be found here:
    #: https://en.wikipedia.org/wiki/List_of_tz_zones_by_name (although not all
    #: systems may support all possibilities). When USE_TZ is True, this is
    #: interpreted as the default user time zone.
    TIME_ZONE: str = 'America/Chicago'

    #: If you set this to True, Django will use timezone-aware datetimes.
    USE_TZ: bool = False

    #: Language code for this installation. All choices can be found here:
    #: http://www.i18nguy.com/unicode/language-identifiers.html
    LANGUAGE_CODE: str = 'en-us'

    #: Languages we provide translations for, out of the box.
    LANGUAGES: List[Tuple] = [
        ('af', gettext_noop('Afrikaans')),
        ('ar', gettext_noop('Arabic')),
        ('ast', gettext_noop('Asturian')),
        ('az', gettext_noop('Azerbaijani')),
        ('bg', gettext_noop('Bulgarian')),
        ('be', gettext_noop('Belarusian')),
        ('bn', gettext_noop('Bengali')),
        ('br', gettext_noop('Breton')),
        ('bs', gettext_noop('Bosnian')),
        ('ca', gettext_noop('Catalan')),
        ('cs', gettext_noop('Czech')),
        ('cy', gettext_noop('Welsh')),
        ('da', gettext_noop('Danish')),
        ('de', gettext_noop('German')),
        ('dsb', gettext_noop('Lower Sorbian')),
        ('el', gettext_noop('Greek')),
        ('en', gettext_noop('English')),
        ('en-au', gettext_noop('Australian English')),
        ('en-gb', gettext_noop('British English')),
        ('eo', gettext_noop('Esperanto')),
        ('es', gettext_noop('Spanish')),
        ('es-ar', gettext_noop('Argentinian Spanish')),
        ('es-co', gettext_noop('Colombian Spanish')),
        ('es-mx', gettext_noop('Mexican Spanish')),
        ('es-ni', gettext_noop('Nicaraguan Spanish')),
        ('es-ve', gettext_noop('Venezuelan Spanish')),
        ('et', gettext_noop('Estonian')),
        ('eu', gettext_noop('Basque')),
        ('fa', gettext_noop('Persian')),
        ('fi', gettext_noop('Finnish')),
        ('fr', gettext_noop('French')),
        ('fy', gettext_noop('Frisian')),
        ('ga', gettext_noop('Irish')),
        ('gd', gettext_noop('Scottish Gaelic')),
        ('gl', gettext_noop('Galician')),
        ('he', gettext_noop('Hebrew')),
        ('hi', gettext_noop('Hindi')),
        ('hr', gettext_noop('Croatian')),
        ('hsb', gettext_noop('Upper Sorbian')),
        ('hu', gettext_noop('Hungarian')),
        ('hy', gettext_noop('Armenian')),
        ('ia', gettext_noop('Interlingua')),
        ('id', gettext_noop('Indonesian')),
        ('io', gettext_noop('Ido')),
        ('is', gettext_noop('Icelandic')),
        ('it', gettext_noop('Italian')),
        ('ja', gettext_noop('Japanese')),
        ('ka', gettext_noop('Georgian')),
        ('kab', gettext_noop('Kabyle')),
        ('kk', gettext_noop('Kazakh')),
        ('km', gettext_noop('Khmer')),
        ('kn', gettext_noop('Kannada')),
        ('ko', gettext_noop('Korean')),
        ('lb', gettext_noop('Luxembourgish')),
        ('lt', gettext_noop('Lithuanian')),
        ('lv', gettext_noop('Latvian')),
        ('mk', gettext_noop('Macedonian')),
        ('ml', gettext_noop('Malayalam')),
        ('mn', gettext_noop('Mongolian')),
        ('mr', gettext_noop('Marathi')),
        ('my', gettext_noop('Burmese')),
        ('nb', gettext_noop('Norwegian Bokmål')),
        ('ne', gettext_noop('Nepali')),
        ('nl', gettext_noop('Dutch')),
        ('nn', gettext_noop('Norwegian Nynorsk')),
        ('os', gettext_noop('Ossetic')),
        ('pa', gettext_noop('Punjabi')),
        ('pl', gettext_noop('Polish')),
        ('pt', gettext_noop('Portuguese')),
        ('pt-br', gettext_noop('Brazilian Portuguese')),
        ('ro', gettext_noop('Romanian')),
        ('ru', gettext_noop('Russian')),
        ('sk', gettext_noop('Slovak')),
        ('sl', gettext_noop('Slovenian')),
        ('sq', gettext_noop('Albanian')),
        ('sr', gettext_noop('Serbian')),
        ('sr-latn', gettext_noop('Serbian Latin')),
        ('sv', gettext_noop('Swedish')),
        ('sw', gettext_noop('Swahili')),
        ('ta', gettext_noop('Tamil')),
        ('te', gettext_noop('Telugu')),
        ('th', gettext_noop('Thai')),
        ('tr', gettext_noop('Turkish')),
        ('tt', gettext_noop('Tatar')),
        ('udm', gettext_noop('Udmurt')),
        ('uk', gettext_noop('Ukrainian')),
        ('ur', gettext_noop('Urdu')),
        ('uz', gettext_noop('Uzbek')),
        ('vi', gettext_noop('Vietnamese')),
        ('zh-hans', gettext_noop('Simplified Chinese')),
        ('zh-hant', gettext_noop('Traditional Chinese')),
    ]

    #: Languages using BiDi (right-to-left) layout
    LANGUAGES_BIDI: List[str] = ["he", "ar", "fa", "ur"]

    #: If you set this to False, Django will make some optimizations so as not
    #: to load the internationalization machinery.
    USE_I18N: bool = True
    LOCALE_PATHS: List = []

    # Settings for language cookie
    LANGUAGE_COOKIE_NAME: str = 'django_language'
    #: The age of the language cookie, in seconds.
    LANGUAGE_COOKIE_AGE: Optional[int] = None
    #: The domain to use for the language cookie.
    #: Set this to a string such as "example.com"
    #: for cross-domain cookies, or use None for a standard domain cookie.
    LANGUAGE_COOKIE_DOMAIN: Optional[str] = None
    LANGUAGE_COOKIE_PATH: str = '/'
    LANGUAGE_COOKIE_SECURE: bool = False
    LANGUAGE_COOKIE_HTTPONLY: bool = False
    LANGUAGE_COOKIE_SAMESITE: Optional[str] = None

    #: If you set this to True, Django will format dates, numbers and calendars
    #: according to user current locale.
    USE_L10N: bool = False

    #: Not-necessarily-technical managers of the site. They get broken link
    #: notifications and other various emails.
    MANAGERS: List[Tuple[str, str]] = ADMINS

    #: Default charset to use for all HttpResponse objects, if a MIME type isn't
    #: manually specified. It's used to construct the Content-Type header.
    DEFAULT_CHARSET: str = 'utf-8'

    #: Encoding of files read from disk (template and initial SQL files).
    FILE_CHARSET: str = 'utf-8'

    #: Email address that error messages come from.
    SERVER_EMAIL: str = 'root@localhost'

    #: Database connection info. If left empty, will default to the dummy backend.
    DATABASES: Dict = {}

    #: Classes used to implement DB routing behavior.
    DATABASE_ROUTERS: list = []

    #: The email backend to use. For possible shortcuts see django.core.mail.
    #: The default is to use the SMTP backend.
    #: Third-party backends can be specified by providing a Python path
    #: to a module that defines an EmailBackend class.
    EMAIL_BACKEND: str = 'django.core.mail.backends.smtp.EmailBackend'

    #: Host for sending email.
    EMAIL_HOST: str = 'localhost'

    #: Port for sending email.
    EMAIL_PORT: int = 25

    #: Whether to send SMTP 'Date' header in the local time zone or in UTC.
    EMAIL_USE_LOCALTIME: bool = False

    #: Optional SMTP authentication information for EMAIL_HOST.
    EMAIL_HOST_USER: str = ''
    EMAIL_HOST_PASSWORD: str = ''
    EMAIL_USE_TLS: bool = False
    EMAIL_USE_SSL: bool = False
    #: If EMAIL_USE_SSL or EMAIL_USE_TLS is True, you can optionally specify
    #: the path to a PEM-formatted certificate chain file to use for the
    #: SSL connection.
    EMAIL_SSL_CERTFILE: Optional[str] = None
    #: If EMAIL_USE_SSL or EMAIL_USE_TLS is True, you can optionally specify
    #: the path to a PEM-formatted private key file to use for the
    #: SSL connection.
    EMAIL_SSL_KEYFILE: Optional[str] = None
    #: Specifies a timeout in seconds for blocking operations like the connection
    #: attempt.
    EMAIL_TIMEOUT: Optional[int] = None

    #: List of strings representing installed apps.
    INSTALLED_APPS: List = []

    TEMPLATES: List = []

    #: Default form rendering class.
    FORM_RENDERER: str = 'django.forms.renderers.DjangoTemplates'

    #: Default email address to use for various automated correspondence from
    #: the site managers.
    DEFAULT_FROM_EMAIL: str = 'webmaster@localhost'

    #: Subject-line prefix for email messages send with django.core.mail.mail_admins
    #: or ...mail_managers.  Make sure to include the trailing space.
    EMAIL_SUBJECT_PREFIX: str = '[Django] '

    #: Whether to append trailing slashes to URLs.
    APPEND_SLASH: bool = True

    #: Whether to prepend the "www." subdomain to URLs that don't have it.
    PREPEND_WWW: bool = False

    #: Override the server-derived value of SCRIPT_NAME
    FORCE_SCRIPT_NAME: Optional[str] = None

    #: List of compiled regular expression objects representing User-Agent strings
    #: that are not allowed to visit any page, systemwide. Use this for bad
    #: robots/crawlers. Here are a few examples:
    #:     import re
    #:     DISALLOWED_USER_AGENTS = [
    #:         re.compile(r'^NaverBot.*'),
    #:         re.compile(r'^EmailSiphon.*'),
    #:         re.compile(r'^SiteSucker.*'),
    #:         re.compile(r'^sohu-search'),
    #:     ]
    DISALLOWED_USER_AGENTS: List = []

    ABSOLUTE_URL_OVERRIDES: Dict = {}

    #: List of compiled regular expression objects representing URLs that need not
    #: be reported by BrokenLinkEmailsMiddleware. Here are a few examples:
    #:    import re
    #:    IGNORABLE_404_URLS = [
    #:        re.compile(r'^/apple-touch-icon.*\.png$'),
    #:        re.compile(r'^/favicon.ico$'),
    #:        re.compile(r'^/robots.txt$'),
    #:        re.compile(r'^/phpmyadmin/'),
    #:        re.compile(r'\.(cgi|php|pl)$'),
    #:    ]
    IGNORABLE_404_URLS: List = []

    #: A secret key for this particular Django installation. Used in secret-key
    #: hashing algorithms. Set this in your settings, or Django will complain
    #: loudly.
    SECRET_KEY: str = ''

    #: Default file storage mechanism that holds media.
    DEFAULT_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'

    #: Absolute filesystem path to the directory that will hold user-uploaded files.
    #: Example: "/var/www/example.com/media/"
    MEDIA_ROOT: str = ''

    #: URL that handles the media served from MEDIA_ROOT.
    #: Examples: "http://example.com/media/", "http://media.example.com/"
    MEDIA_URL: str = ''

    #: Absolute path to the directory static files should be collected to.
    #: Example: "/var/www/example.com/static/"
    STATIC_ROOT: Optional[str] = None

    #: URL that handles the static files served from STATIC_ROOT.
    #: Example: "http://example.com/static/", "http://static.example.com/"
    STATIC_URL: Optional[str] = None

    #: List of upload handler classes to be applied in order.
    FILE_UPLOAD_HANDLERS: List[str] = [
        'django.core.files.uploadhandler.MemoryFileUploadHandler',
        'django.core.files.uploadhandler.TemporaryFileUploadHandler',
    ]

    #: Maximum size, in bytes, of a request before it will be streamed to the
    #: file system instead of into memory.
    FILE_UPLOAD_MAX_MEMORY_SIZE: int = 2621440  # i.e. 2.5 MB

    #: Maximum size in bytes of request data (excluding file uploads) that will be
    #: read before a SuspiciousOperation (RequestDataTooBig) is raised.
    DATA_UPLOAD_MAX_MEMORY_SIZE: int = 2621440  # i.e. 2.5 MB

    #: Maximum number of GET/POST parameters that will be read before a
    #: SuspiciousOperation (TooManyFieldsSent) is raised.
    DATA_UPLOAD_MAX_NUMBER_FIELDS: int = 1000

    #: Directory in which upload streamed files will be temporarily saved. A value of
    #: `None` will make Django use the operating system's default temporary directory
    #: (i.e. "/tmp" on *nix systems).
    FILE_UPLOAD_TEMP_DIR: Optional[str] = None

    #: The numeric mode to set newly-uploaded files to. The value should be a mode
    #: you'd pass directly to os.chmod;
    # see https://docs.python.org/library/os.html#files-and-directories.
    FILE_UPLOAD_PERMISSIONS: int = 0o644

    #: The numeric mode to assign to newly-created directories, when uploading files.
    #: The value should be a mode as you'd pass to os.chmod;
    #: see https://docs.python.org/library/os.html#files-and-directories.
    FILE_UPLOAD_DIRECTORY_PERMISSIONS: Optional[int] = None

    #: Python module path where user will place custom format definition.
    #: The directory where this setting is pointing should contain subdirectories
    #: named as the locales, containing a formats.py file
    #: (i.e. "myproject.locale" for myproject/locale/en/formats.py etc. use)
    FORMAT_MODULE_PATH: Optional[str] = None

    #: Default formatting for date objects. See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    DATE_FORMAT: str = 'N j, Y'

    #: Default formatting for datetime objects. See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    DATETIME_FORMAT: str = 'N j, Y, P'

    #: Default formatting for time objects. See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    TIME_FORMAT: str = 'P'

    #: Default formatting for date objects when only the year and month are relevant.
    #: See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    YEAR_MONTH_FORMAT: str = 'F Y'

    #: Default formatting for date objects when only the month and day are relevant.
    #: See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    MONTH_DAY_FORMAT: str = 'F j'

    #: Default short formatting for date objects. See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    SHORT_DATE_FORMAT: str = 'm/d/Y'

    #: Default short formatting for datetime objects.
    #: See all available format strings here:
    #: https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
    SHORT_DATETIME_FORMAT: str = 'm/d/Y P'

    #: Default formats to be used when parsing dates from input boxes, in order
    #: See all available format string here:
    #: https://docs.python.org/library/datetime.html#strftime-behavior
    #: * Note that these format strings are different from the ones to display dates
    DATE_INPUT_FORMATS: List[str] = [
        '%Y-%m-%d',
        '%m/%d/%Y',
        '%m/%d/%y',  # '2006-10-25', '10/25/2006', '10/25/06'
        '%b %d %Y',
        '%b %d, %Y',  # 'Oct 25 2006', 'Oct 25, 2006'
        '%d %b %Y',
        '%d %b, %Y',  # '25 Oct 2006', '25 Oct, 2006'
        '%B %d %Y',
        '%B %d, %Y',  # 'October 25 2006', 'October 25, 2006'
        '%d %B %Y',
        '%d %B, %Y',  # '25 October 2006', '25 October, 2006'
    ]

    #: Default formats to be used when parsing times from input boxes, in order
    #: See all available format string here:
    #: https://docs.python.org/library/datetime.html#strftime-behavior
    #: * Note that these format strings are different from the ones to display dates
    TIME_INPUT_FORMATS: List[str] = [
        '%H:%M:%S',  # '14:30:59'
        '%H:%M:%S.%f',  # '14:30:59.000200'
        '%H:%M',  # '14:30'
    ]

    #: Default formats to be used when parsing dates and times from input boxes,
    #: in order
    #: See all available format string here:
    #: https://docs.python.org/library/datetime.html#strftime-behavior
    #: * Note that these format strings are different from the ones to display dates
    DATETIME_INPUT_FORMATS: List[str] = [
        '%Y-%m-%d %H:%M:%S',  # '2006-10-25 14:30:59'
        '%Y-%m-%d %H:%M:%S.%f',  # '2006-10-25 14:30:59.000200'
        '%Y-%m-%d %H:%M',  # '2006-10-25 14:30'
        '%Y-%m-%d',  # '2006-10-25'
        '%m/%d/%Y %H:%M:%S',  # '10/25/2006 14:30:59'
        '%m/%d/%Y %H:%M:%S.%f',  # '10/25/2006 14:30:59.000200'
        '%m/%d/%Y %H:%M',  # '10/25/2006 14:30'
        '%m/%d/%Y',  # '10/25/2006'
        '%m/%d/%y %H:%M:%S',  # '10/25/06 14:30:59'
        '%m/%d/%y %H:%M:%S.%f',  # '10/25/06 14:30:59.000200'
        '%m/%d/%y %H:%M',  # '10/25/06 14:30'
        '%m/%d/%y',  # '10/25/06'
    ]

    #: First day of week, to be used on calendars
    #: 0 means Sunday, 1 means Monday...
    FIRST_DAY_OF_WEEK: int = 0

    #: Decimal separator symbol
    DECIMAL_SEPARATOR: str = '.'

    #: Boolean that sets whether to add thousand separator when formatting numbers
    USE_THOUSAND_SEPARATOR: bool = False

    #: Number of digits that will be together, when splitting them by
    #: THOUSAND_SEPARATOR. 0 means no grouping, 3 means splitting by thousands...
    NUMBER_GROUPING: int = 0

    #: Thousand separator symbol
    THOUSAND_SEPARATOR: str = ','

    #: The tablespaces to use for each model when not specified otherwise.
    DEFAULT_TABLESPACE: str = ''
    DEFAULT_INDEX_TABLESPACE: str = ''

    #: Default X-Frame-Options header value
    X_FRAME_OPTIONS: str = 'DENY'

    USE_X_FORWARDED_HOST: bool = False
    USE_X_FORWARDED_PORT: bool = False

    #: The Python dotted path to the WSGI application that Django's internal server
    #: (runserver) will use. If `None`, the return value of
    #: 'django.core.wsgi.get_wsgi_application' is used, thus preserving the same
    #: behavior as previous versions of Django. Otherwise this should point to an
    #: actual WSGI application object.
    WSGI_APPLICATION: Optional[str] = None

    #: If your Django app is behind a proxy that sets a header to specify secure
    #: connections, AND that proxy ensures that user-submitted headers with the
    #: same name are ignored (so that people can't spoof it), set this value to
    #: a tuple of (header_name, header_value). For any requests that come in with
    #: that header/value, request.is_secure() will return True.
    #: WARNING! Only set this if you fully understand what you're doing. Otherwise,
    #: you may be opening yourself up to a security risk.
    SECURE_PROXY_SSL_HEADER: Optional[Tuple] = None

    ##############
    # MIDDLEWARE #
    ##############

    # List of middleware to use. Order is important; in the request phase, these
    # middleware will be applied in the order given, and in the response
    # phase the middleware will be applied in reverse order.
    MIDDLEWARE: List[str] = []

    ############
    # SESSIONS #
    ############

    #: Cache to store session data if using the cache session backend.
    SESSION_CACHE_ALIAS: str = 'default'
    #: Cookie name. This can be whatever you want.
    SESSION_COOKIE_NAME: str = 'sessionid'
    #: Age of cookie, in seconds (default: 2 weeks).
    SESSION_COOKIE_AGE: int = 60 * 60 * 24 * 7 * 2
    #: A string like "example.com", or None for standard domain cookie.
    SESSION_COOKIE_DOMAIN: Optional[str] = None
    #: Whether the session cookie should be secure (https:// only).
    SESSION_COOKIE_SECURE: bool = False
    #: The path of the session cookie.
    SESSION_COOKIE_PATH: str = '/'
    #: Whether to use the HttpOnly flag.
    SESSION_COOKIE_HTTPONLY: bool = True
    #: Whether to set the flag restricting cookie leaks on cross-site requests.
    #: This can be 'Lax', 'Strict', or None to disable the flag.
    SESSION_COOKIE_SAMESITE: Optional[str] = 'Lax'
    #: Whether to save the session data on every request.
    SESSION_SAVE_EVERY_REQUEST: bool = False
    #: Whether a user's session cookie expires when the Web browser is closed.
    SESSION_EXPIRE_AT_BROWSER_CLOSE: bool = False
    #: The module to store session data
    SESSION_ENGINE: str = 'django.contrib.sessions.backends.db'
    #: Directory to store session files if using the file session module. If None,
    #: the backend will use a sensible default.
    SESSION_FILE_PATH: Optional[str] = None
    #: class to serialize session data
    SESSION_SERIALIZER: str = 'django.contrib.sessions.serializers.JSONSerializer'

    #########
    # CACHE #
    #########

    # The cache backends to use.
    CACHES: dict = {
        'default': {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}
    }
    CACHE_MIDDLEWARE_KEY_PREFIX: str = ''
    CACHE_MIDDLEWARE_SECONDS: int = 600
    CACHE_MIDDLEWARE_ALIAS: str = 'default'

    ##################
    # AUTHENTICATION #
    ##################

    AUTH_USER_MODEL: str = 'auth.User'

    AUTHENTICATION_BACKENDS: List[str] = ['django.contrib.auth.backends.ModelBackend']

    LOGIN_URL: str = '/accounts/login/'

    LOGIN_REDIRECT_URL: str = '/accounts/profile/'

    LOGOUT_REDIRECT_URL: Optional[str] = None

    #: The number of days a password reset link is valid for
    PASSWORD_RESET_TIMEOUT_DAYS: int = 3

    #: The first hasher in this list is the preferred algorithm.  any
    #: password using different algorithms will be converted automatically
    #: upon login
    PASSWORD_HASHERS: List[str] = [
        'django.contrib.auth.hashers.PBKDF2PasswordHasher',
        'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
        'django.contrib.auth.hashers.Argon2PasswordHasher',
        'django.contrib.auth.hashers.BCryptSHA256PasswordHasher',
    ]

    AUTH_PASSWORD_VALIDATORS: List[Dict] = []

    ###########
    # SIGNING #
    ###########

    SIGNING_BACKEND: str = 'django.core.signing.TimestampSigner'

    ########
    # CSRF #
    ########

    #: Dotted path to callable to be used as view when a request is
    #: rejected by the CSRF middleware.
    CSRF_FAILURE_VIEW: str = 'django.views.csrf.csrf_failure'

    # Settings for CSRF cookie.
    CSRF_COOKIE_NAME: str = 'csrftoken'
    CSRF_COOKIE_AGE: int = 60 * 60 * 24 * 7 * 52
    CSRF_COOKIE_DOMAIN: Optional[str] = None
    CSRF_COOKIE_PATH: str = '/'
    CSRF_COOKIE_SECURE: bool = False
    CSRF_COOKIE_HTTPONLY: bool = False
    CSRF_COOKIE_SAMESITE: Optional[str] = 'Lax'
    CSRF_HEADER_NAME: str = 'HTTP_X_CSRFTOKEN'
    CSRF_TRUSTED_ORIGINS: List[str] = []
    CSRF_USE_SESSIONS: bool = False

    ############
    # MESSAGES #
    ############

    #: Class to use as messages backend
    MESSAGE_STORAGE: str = 'django.contrib.messages.storage.fallback.FallbackStorage'

    # Default values of MESSAGE_LEVEL and MESSAGE_TAGS are defined within
    # django.contrib.messages to avoid imports in this settings file.

    ###########
    # LOGGING #
    ###########

    #: The callable to use to configure logging
    LOGGING_CONFIG: str = 'logging.config.dictConfig'

    #: Custom logging configuration.
    LOGGING: Dict = {}

    #: Default exception reporter filter class used in case none has been
    #: specifically assigned to the HttpRequest instance.
    DEFAULT_EXCEPTION_REPORTER_FILTER: str = 'django.views.debug.SafeExceptionReporterFilter'

    ###########
    # TESTING #
    ###########

    #: The name of the class to use to run the test suite
    TEST_RUNNER: str = 'django.test.runner.DiscoverRunner'

    #: Apps that don't need to be serialized at test database creation time
    #: (only apps with migrations are to start with)
    TEST_NON_SERIALIZED_APPS: List[str] = []

    ############
    # FIXTURES #
    ############

    #: The list of directories to search for fixtures
    FIXTURE_DIRS: List[str] = []

    ###############
    # STATICFILES #
    ###############

    #: A list of locations of additional static files
    STATICFILES_DIRS: List[str] = []

    #: The default file storage backend used during the build process
    STATICFILES_STORAGE: str = 'django.contrib.staticfiles.storage.StaticFilesStorage'

    #: List of finder classes that know how to find static files in
    #: various locations.
    STATICFILES_FINDERS: List[str] = [
        'django.contrib.staticfiles.finders.FileSystemFinder',
        'django.contrib.staticfiles.finders.AppDirectoriesFinder',
        # 'django.contrib.staticfiles.finders.DefaultStorageFinder',
    ]

    ##############
    # MIGRATIONS #
    ##############

    #: Migration module overrides for apps, by app label.
    MIGRATION_MODULES: Dict[str, str] = {}

    #################
    # SYSTEM CHECKS #
    #################

    #: List of all issues generated by system checks that should be silenced. Light
    #: issues like warnings, infos or debugs will not generate a message. Silencing
    #: serious issues like errors and criticals does not result in hiding the
    #: message, but Django will not stop you from e.g. running server.
    SILENCED_SYSTEM_CHECKS: List[str] = []

    #######################
    # SECURITY MIDDLEWARE #
    #######################
    SECURE_BROWSER_XSS_FILTER: bool = False
    SECURE_CONTENT_TYPE_NOSNIFF: bool = True
    SECURE_HSTS_INCLUDE_SUBDOMAINS: bool = False
    SECURE_HSTS_PRELOAD: bool = False
    SECURE_HSTS_SECONDS: int = 0
    SECURE_REDIRECT_EXEMPT: List[str] = []
    SECURE_REFERRER_POLICY: Optional[str] = None
    SECURE_SSL_HOST: Optional[str] = None
    SECURE_SSL_REDIRECT: bool = False

    ##############
    # Undefined
    ##############

    #: A string representing the full Python import path to your root URLconf,
    #: for example "mydjangoapps.urls".
    ROOT_URLCONF: str = Undefined @ required



import os
from typing import Any, Tuple, Optional, Union, Type

from concrete_settings.sources import (
    Source,
    StringSourceMixin,
    AnySource,
    register_source,
    NotFound)


@register_source
class EnvVarSource(StringSourceMixin, Source):
    def __init__(self):
        self.data = os.environ

    @staticmethod
    def get_source(src: AnySource) -> Optional['EnvVarSource']:
        if isinstance(src, EnvVarSource):
            return src
        else:
            return None

    def read(self, setting, parents: Tuple[str, ...] = ()) -> Union[Type[NotFound], Any]:
        parents_upper = map(str.upper, parents)
        key = '_'.join((*parents_upper, setting.name))
        val = os.environ.get(key)

        if val is None:
            return NotFound
        else:
            return self.convert_value(val, setting.type_hint)



from .behavior import Behavior, GetterSetterBehavior  # noqa: F401 # imported but unused
from .override import override  # noqa: F401 # imported but unused
from .validate import validate  # noqa: F401 # imported but unused



from pathlib import Path
from typing import Any, Dict, List, Tuple, Type, Optional, Union
from typing import TYPE_CHECKING

from ..exceptions import ConcreteSettingsError
from . import strategies  # noqa: F401 # imported but unused

if TYPE_CHECKING:
    from ..settings import Setting


_registered_sources = set()

AnySource = Union[Dict[str, Any], str, 'Source', Path]


class NotFound:
    pass


def register_source(source_cls: Type['Source']):
    global _registered_sources
    _registered_sources.add(source_cls)
    return source_cls


class NoSuitableSourceFound(ConcreteSettingsError):
    def __init__(self, src: AnySource):
        super().__init__(
            f'No suitable source found to handle "{src}".\n'
            'Perhaps you have forgotten to register the source?'
        )


def get_source(src: AnySource) -> 'Source':
    if isinstance(src, Source):
        return src

    for src_cls in _registered_sources:
        source = src_cls.get_source(src)
        if source is not None:
            return source

    raise NoSuitableSourceFound(src)


class Source:
    @staticmethod
    def get_source(src: AnySource) -> Optional['Source']:
        return None

    def read(
        self, setting: 'Setting', parents: Tuple[str, ...] = ()
    ) -> Union[Type[NotFound], Any]:
        pass


class StringSourceMixin:
    """Extends source by providing a string value to required type
       conversion method."""

    @staticmethod
    def convert_value(val: str, type_hint: Any = None) -> Any:
        """Convert given string value to type based on `type_hint`"""
        if type_hint in (int, float):
            return type_hint(val)
        elif type_hint is bool:
            if val.lower() == 'true':
                return True
            elif val.lower() == 'false':
                return False

        return val


@register_source
class DictSource(Source):
    def __init__(self, s: dict):
        self.data: dict = s

    @staticmethod
    def get_source(src: AnySource) -> Optional['DictSource']:
        if isinstance(src, dict):
            return DictSource(src)
        elif isinstance(src, DictSource):
            return src
        else:
            return None

    def read(
        self, setting: 'Setting', parents: Tuple[str, ...] = ()
    ) -> Union[Type[NotFound], Any]:
        d = self.data
        for key in parents:
            d = d[key]

        val = d.get(setting.name, NotFound)
        return val


class FileSource(Source):
    extensions: List[str] = []

    path: str

    def __init__(self, path):
        self.path = path

    @classmethod
    def get_source(cls, src) -> Optional['FileSource']:
        if isinstance(src, cls):
            return src

        if isinstance(src, Path):
            src = str(src.resolve())

        if isinstance(src, str):
            for ext in cls.extensions:
                if src.endswith(ext):
                    return cls(src)

        return None



from concrete_settings.behaviors import Behavior
from concrete_settings.validators import RequiredValidator


class required(Behavior):
    def __init__(self, message: str = None):
        self.message = message

    def decorate(self, setting):
        setting.validators = (RequiredValidator(self.message),) + setting.validators
        super().decorate(setting)



from typing import Union, List, Dict


class ConcreteSettingsError(Exception):
    """Base class for all concrete_settings exceptions."""


class StructureError(ConcreteSettingsError):
    """Raised when an inconsistency in settings inheritance hierarchy is detected."""


SettingName = str

# fmt: off
# type ignored due to cyclic definition, promised to be fixed in
# future mypy versions

#:
ValidationErrorDetails = Union[                   # type: ignore
    str,
    List['ValidationErrorDetails'],               # type: ignore
    Dict[SettingName, 'ValidationErrorDetails'],  # type: ignore
]
# fmt: on


class ValidationError(ConcreteSettingsError):
    sources: List[str]

    def __init__(self, details: ValidationErrorDetails = ''):
        self.details = details
        self.sources = []

    def prepend_source(self, source):
        self.sources.insert(0, source)

    def __str__(self):
        detail_str = _format_detail(self.details)
        if self.sources:
            source = '.'.join(self.sources)
            return f'{source}: {detail_str}'
        else:
            return detail_str


def _format_detail(detail) -> str:
    if isinstance(detail, list):
        return '; '.join(_format_detail(d) for d in detail)
    elif isinstance(detail, dict):
        return '\n'.join(f'{k}: {_format_detail(v)}.' for k, v in detail.items())
    else:
        return str(detail)



import warnings

from concrete_settings.exceptions import ValidationError
from concrete_settings.validators import Validator


class DeprecatedValidator(Validator):
    def __init__(self, msg, raise_exception: bool = False):
        self.msg = msg
        self.raise_exception = raise_exception

    def __call__(self, value, *, name, owner, **ignore):
        msg = self.msg.format(name=name, owner=type(owner))

        if self.raise_exception:
            raise ValidationError(msg)
        else:
            warnings.warn(msg, DeprecationWarning)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['concrete_settings',
 'concrete_settings.behaviors',
 'concrete_settings.contrib',
 'concrete_settings.contrib.behaviors',
 'concrete_settings.contrib.frameworks',
 'concrete_settings.contrib.sources',
 'concrete_settings.contrib.validators',
 'concrete_settings.sources',
 'concrete_settings.validators']

package_data = \
{'': ['*']}

install_requires = \
['sphinx>=3.0,<4.0', 'typeguard>=2.7,<3.0', 'typing_extensions>=3.7.4,<4.0.0']

extras_require = \
{'pyyaml': ['pyyaml>=5.3,<6.0']}

setup_kwargs = {
    'name': 'concrete-settings',
    'version': '0.1',
    'description': 'Concrete Settings facilitates configuration management in big and small projects.',
    'long_description': 'Concrete Settings\n#################\n\n.. image:: https://travis-ci.org/BasicWolf/concrete-settings.svg?branch=master\n    :target: https://travis-ci.org/BasicWolf/concrete-settings\n\n.. image:: https://readthedocs.org/projects/concrete-settings/badge/?version=latest\n   :target: https://concrete-settings.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://raw.githubusercontent.com/BasicWolf/concrete-settings/master/docs/src/_static/img/codestyle_black.svg\n    :target: https://github.com/ambv/black\n\n.. image:: https://raw.githubusercontent.com/BasicWolf/concrete-settings/master/docs/src/_static/img/mypy_checked.svg\n   :target: https://github.com/python/mypy\n\n.. image:: https://raw.githubusercontent.com/BasicWolf/concrete-settings/master/docs/src/_static/img/pyup_scanned.svg\n   :target: https://pyup.io\n\n\nWelcome to Concrete Settings\n============================\n\n**Concrete Settings** is a Python library which facilitates\nconfiguration management in big and small projects.\n\nThe library was born out of necessity to manage a huge\ndecade-old Django-based SaaS solution with more than two hundred\ndifferent application settings scattered around ``settings.py``.\n*What does this setting do?*\n*What type is it?*\n*Why does it have such a weird format?*\n*Is this the final value, or it changes somewhere on the way?*\nSometimes developers spent *hours* seeking answers to these\nquestions.\n\nConcrete Settigns tackles these problems altogether.\nIt was designed to be developer and end-user friendly.\nThe settings are defined via normal Python code with few\ntricks which significantly improve readability\nand maintainability.\n\n.. contents:: :depth: 2\n\nA small example\n---------------\n\nTake a look at a small example of Settings class with one\nboolean setting ``DEBUG``. A developer defines the\nsettings in application code, while an end-user\nchooses to store the configuration in a YAML file:\n\n.. code-block:: python\n\n   # settings.py\n\n   from concrete_settings import Settings\n\n   class AppSettings(Settings):\n\n       #: Turns debug mode on/off\n       DEBUG: bool = False\n\n\n   app_settings = AppSettings()\n   app_settings.update(\'/path/to/user/settings.yml\')\n   app_settings.is_valid(raise_exception=True)\n\n.. code-block:: yaml\n\n   # settings.yml\n\n   DEBUG: true\n\n\nAccessing settings:\n\n.. code-block:: pycon\n\n   >>>  print(app_settings.DEBUG)\n   True\n\n   >>> print(AppSettings.DEBUG.__doc__)\n   Turns debug mode on/off\n\nConcrete concepts\n-----------------\n\nSettings are **defined in classes**. Python mechanism\nof inheritance and composition apply here, so settings can be **mixed** (multiple inheritance)\nand be **nested** (settings as class fields).\nSettings are **type-annotated** and are **validated**.\nDocumentation matters! Each settings can be documented in Sphinx-style comments ``#:`` written\nabove its definition.\nAn instance of ``Settings`` can be updated i.e. read from any kind of source:\nYAML, JSON or Python files, environmental variables, Python dicts, and you can add more!\n\nFinally, Concrete Settings comes with batteries like Django 3.0 support out of the box.\n\nConcrete Settings are here to improve configuration handling\nwhether you are starting from scratch, or dealing with an\nold legacy Cthulhu.\nAre you ready to try it out?\n\n``pip install concrete-settings`` and welcome to the `documentation <https://concrete-settings.readthedocs.io/>`__!\n\nWhat\'s inside?\n==============\n\nSo, you are a kind of a developer who expects more show cases in a ``README``?\nLet\'s see!\n\nCatch an invalid value early\n----------------------------\n\nFor example, the default *type validator* works like this:\n\n.. code-block:: python\n\n   from concrete_settings import Settings\n\n   class AppSettings(Settings):\n       SPEED: int = \'abc\'\n\n   app_settings = AppSettings()\n   print(app_settings.is_valid(raise_exception=False))\n   print(app_settings.errors)\n\nOutput:\n\n.. code-block::\n\n   False\n   {\'SPEED\': ["Expected value of type `<class \'int\'>` got value of type `<class \'str\'>`"]}\n\n\nEasily warn about deprecation\n-----------------------------\n\nUse **behaviors** to control settings during their *initialization*, *validation*,\n*reading* and *writing* operations:\n\n.. code-block:: python\n\n   from concrete_settings import Settings, Setting\n   from concrete_settings.contrib.behaviors import deprecated\n\n   class AppSettings(Settings):\n       MAX_SPEED: int = 30 @deprecated\n\n   app_settings = AppSettings()\n   app_settings.is_valid()\n\nRunning this code with ``python -Wdefault`` yields:\n\n.. code-block::\n\n   DeprecationWarning: Setting `MAX_SPEED` in class `<class \'__main__.AppSettings\'>` is deprecated.\n\n\nGroup settings and nest them\n----------------------------\n\nDifferent settings in a huge setting file?\nWhy have those stupid ``GROUP_PREFIXES_...``?\nInstead group and nest settings:\n\n.. code-block:: python\n\n   from concrete_settings import Settings\n\n   class DBSettings(Settings):\n       USER = \'alex\'\n       PASSWORD  = \'secret\'\n       SERVER = \'localhost@5432\'\n\n   class CacheSettings(Settings):\n       ENGINE = \'DatabaseCache\'\n       TIMEOUT = 300\n\n   class LoggingSettings(Settings):\n       LEVEL = \'INFO\'\n       FORMAT = \'%(asctime)s %(levelname)-8s %(name)-15s %(message)s\'\n\n\n   class AppSettings(Settings):\n       DB = DBSettings()\n       CACHE = CacheSettings()\n       LOG = LoggingSettings()\n\n   app_settings = AppSettings()\n   app_settings.is_valid()  # also invokes DB, CACHE and LOG validation\n   print(app_settings.LOG.LEVEL)\n\nThere is even more\n------------------\n\nThere is even more for you to discover in\n`documentation <https://basicwolf.github.io/concrete-settings>`__,\nand there is more to come. **Your** contribution, be it\na *bug report*, *pull request*, *suggested feature*,\n*comments* and *criticism* are very welcome!\n\nAwesome configuration projects\n==============================\n\nConcrete Settings is not the first and surely not the last library to handle\nconfiguration in Python projects.\n\n* `goodconf <https://github.com/lincolnloop/goodconf/>`_\n  - Define configuration variables and load them from environment or\n  JSON/YAML file. Also generates initial configuration files and\n  documentation for your defined configuration.\n\n* `profig <https://profig.readthedocs.io>`_\n  - is a straightforward configuration library for Python.\n  Its objective is to make the most common tasks of configuration\n  handling as simple as possible.\n\n* `everett <https://everett.readthedocs.io/en/latest/>`_\n  - is a Python configuration library with the following goals:\n  flexible configuration from multiple configured environments;\n  easy testing with configuration and easy documentation of configuration\n  for users.\n\n* `python-decouple <https://github.com/henriquebastos/python-decouple>`_\n  - strict separation of settings from code. Decouple helps you to organize\n  your settings so that you can change parameters without having to redeploy\n  your app.\n\nWhy should you trust Concrete Settings instead of picking some other library?\nConcrete Settings tries to make configuration definition, processing and maintenance smooth and transparent for developers. Its implicit definition syntax eliminates extra code and allows you to focus on  what is important.\n\n\nWords of gratitude\n==================\n\nIt is hard to imagine a software project without the infrastructure that supports it.\nConcrete Settings exists as an open source library only thanks to the services and the tools\nthat are used to host, build and maintain it.\n\nMy warmest words of gratitude go to\n`Github <https://github.com>`_ for hosting the project source,\n`Read the Docs <https://readthedocs.org>`_ for hosting the documentation,\n`Travis CI <https://travis-ci.org>`_ for continuous integration process\nand`pyupio <https://pyup.io>`_ for keeping dependencies up-to-date!\n\nMany thanks for the authors and contributors of the libraries used to build the project.\n\nAll these wonderful services and tools have been provided *pro bono*.\n',
    'author': 'Zaur Nasibov',
    'author_email': 'comments@znasibov.info',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/basicwolf/concrete-settings',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import json

from typing import Any, Tuple, Union, Type

from concrete_settings.exceptions import ConcreteSettingsError
from concrete_settings.sources import FileSource, register_source, NotFound


@register_source
class JsonSource(FileSource):
    extensions = ['.json', '.js']

    def __init__(self, path):
        super().__init__(path)
        self._data = None

    def read(
        self, setting, parents: Tuple[str, ...] = ()
    ) -> Union[Type[NotFound], Any]:
        if self._data is None:
            self._data = self._read_file(self.path)

        d = self._data
        for key in parents:
            d = d[key]

        val = d.get(setting.name, NotFound)
        return val

    @staticmethod
    def _read_file(path):
        try:
            with open(path) as f:
                raw_data = f.read()
                return json.loads(raw_data)
        except FileNotFoundError as e:
            raise ConcreteSettingsError(f"Source file {path} was not found") from e
        except json.decoder.JSONDecodeError as e:
            raise ConcreteSettingsError(f"Error parsing JSON from {path}: {e}") from e



from typing import TYPE_CHECKING

from .behavior import Behavior

if TYPE_CHECKING:
    from ..settings import Setting


class override(Behavior):
    def decorate(self, setting: 'Setting'):
        setting.override = True
        super().decorate(setting)



import warnings
from typing import TYPE_CHECKING, Any

from concrete_settings.behaviors.behavior import GetterSetterBehavior
from concrete_settings.contrib.validators import DeprecatedValidator

if TYPE_CHECKING:
    from concrete_settings import Setting, Settings


class deprecated(GetterSetterBehavior):
    def __init__(
        self,
        deprecation_message: str = 'Setting `{name}` in class `{owner}` is deprecated.',
        *,
        warn_on_validation=True,
        error_on_validation=False,
        warn_on_get=False,
        warn_on_set=False,
    ):
        self.deprecation_message = deprecation_message
        self.error_on_validation = error_on_validation
        self.warn_on_validation = warn_on_validation
        self.warn_on_get = warn_on_get
        self.warn_on_set = warn_on_set

    def decorate(self, setting: 'Setting'):
        if self.warn_on_validation or self.error_on_validation:
            setting.validators = (
                DeprecatedValidator(self.deprecation_message, self.error_on_validation),
            ) + setting.validators
        super().decorate(setting)

    def get_value(self, setting: 'Setting', owner: 'Settings') -> Any:
        if self.warn_on_get:
            if owner and not owner.is_being_validated:
                msg = self.deprecation_message.format(owner=owner, name=setting.name)
                warnings.warn(msg, DeprecationWarning)
        return super().get_value(setting, owner)

    def set_value(self, setting: 'Setting', owner: 'Settings', value):
        if self.warn_on_set:
            if not owner.is_being_validated:
                msg = self.deprecation_message.format(owner=owner, name=setting.name)
                warnings.warn(msg, DeprecationWarning)
        super().set_value(setting, owner, value)



from typing import TYPE_CHECKING, Tuple

from .behavior import Behavior, BehaviorWithArgumentsMeta

if TYPE_CHECKING:
    from ..settings import Setting
    from ..validators import Validator


class validate(Behavior, metaclass=BehaviorWithArgumentsMeta):
    def __init__(self, *validators: 'Validator'):
        self._validators: Tuple['Validator', ...] = validators

    def decorate(self, setting: 'Setting'):
        setting.validators += self._validators
        super().decorate(setting)



import sys

from . import settings  # noqa: F401 # imported but unused
from .settings import (  # noqa: F401 # imported but unused
    Settings,
    Setting,
    PropertySetting,
    INVALID_SETTINGS,
)

from .exceptions import ValidationError  # noqa: F401 # imported but unused
from .validators import Validator  # noqa: F401 # imported but unused
from .types import Undefined  # noqa: F401 # imported but unused
from .sources import register_source  # noqa: F401 # imported but unused
from .contrib.behaviors import required  # noqa: F401 # imported but unused
from .behaviors import (  # noqa: F401 # imported but unused
    Behavior,
    GetterSetterBehavior,
    override,
    validate,
)

setting = PropertySetting

name = "concrete_settings"
PY_VERSION = (sys.version_info.major, sys.version_info.minor)
PY_36 = (3, 6)

if PY_VERSION < PY_36:
    raise ImportError("Python 3.6 or higher is required by concrete_settings library")


def _load_contrib():
    from .contrib.sources import YamlSource, JsonSource, EnvVarSource  # noqa: F401


_load_contrib()


__version__ = "0.1"



from typeguard import check_type

from concrete_settings.exceptions import ValidationError
from concrete_settings.types import Undefined
from .validator import Validator


class ValueTypeValidator(Validator):
    def __init__(self, type_hint=None):
        self.type_hint = type_hint

    def __call__(self, value, *, name, setting, **ignore):
        if value is Undefined:
            return

        type_hint = setting.type_hint if self.type_hint is None else self.type_hint

        try:
            check_type(name, value, type_hint)
        except TypeError as e:
            raise ValidationError(
                f'Expected value of type `{type_hint}` '
                f'got value of type `{type(value)}`'
            ) from e



# Concrete Settings Django 3.0 project settings template

# Remember to fill the placeholders:
# * secret_key
# * mysite.urls
# * mysite.wsgi.application


# flake8: noqa
# type: ignore

import os

from concrete_settings.contrib.frameworks.django30 import Django30Settings

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/


class MyProjectSettings(Django30Settings):
    # SECURITY WARNING: keep the secret key used in production secret!
    SECRET_KEY = 'secret_key'

    # SECURITY WARNING: don't run with debug turned on in production!
    DEBUG = True

    ALLOWED_HOSTS = []

    # Application definition

    INSTALLED_APPS = [
        'django.contrib.admin',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.messages',
        'django.contrib.staticfiles',
    ]

    MIDDLEWARE = [
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    ]

    ROOT_URLCONF = 'mysite.urls'

    TEMPLATES = [
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                ],
            },
        },
    ]

    WSGI_APPLICATION = 'mysite.wsgi.application'

    # Database
    # https://docs.djangoproject.com/en/3.0/ref/settings/#databases

    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': '/path/to/db.sqlite3',  # e.g. os.path.join(BASE_DIR, 'db.sqlite3'),
        }
    }

    # Password validation
    # https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

    AUTH_PASSWORD_VALIDATORS = [
        {
            'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
        },
        {'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator', },
        {'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator', },
        {'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator', },
    ]

    # Internationalization
    # https://docs.djangoproject.com/en/3.0/topics/i18n/

    LANGUAGE_CODE = 'en-us'

    TIME_ZONE = 'UTC'

    USE_I18N = True

    USE_L10N = True

    USE_TZ = True

    # Static files (CSS, JavaScript, Images)
    # https://docs.djangoproject.com/en/3.0/howto/static-files/
    STATIC_URL = '/static/'
